from ._csv_file_layouts import get_format, ColumnParseParameters, spectral_column_names
from ._spotter_constants import spotter_constants, SpotterConstants
from pandas import DataFrame, read_csv, concat, to_datetime
from roguewave.timeseries_analysis.filtering import sos_filter
from typing import Iterator, List, Callable
from xarray import Dataset
from glob import glob
from roguewave import FrequencySpectrum
from roguewave.tools.time import datetime64_to_timestamp, to_datetime_utc, to_datetime64
from datetime import datetime
import os
from numpy import (
    nan,
    linspace,
    errstate,
    sqrt,
    interp,
    full_like,
    inf,
    cos,
    sin,
    pi,
    timedelta64,
)

_pattern = {
    "GPS": "????_GPS.csv",
    "FLT": "????_FLT.csv",
    "LOC": "????_LOC.csv",
    "SPC": "????_SPC.csv",
    "TIME": "????_LOC.csv",
}


def apply_to_group(function: Callable[[DataFrame], DataFrame], dataframe: DataFrame):
    """
    Apply a function to each group seperately and recombine the result into a single dataframe

    :param function: Function to appply
    :param dataframe: Dataframe to apply function to
    :return:
    """
    groups = dataframe.groupby("group id")
    dataframes = [function(x[1]) for x in groups]
    return concat(dataframes)


def files_to_parse(
    path: str, pattern: str, start_date: datetime = None, end_date: datetime = None
) -> Iterator[str]:

    files = glob(os.path.join(path, pattern))

    if (start_date is not None) or (end_date is not None):
        time_files = glob(os.path.join(path, _pattern["TIME"]))
        csv_parsing_options = get_format("TIME")
        if not (len(time_files) == len(files)):
            raise ValueError(
                "Number of location files needs to be the same as the number of files being read to use"
                "time filtering"
            )

        for file, time_file in zip(sorted(files), sorted(time_files)):
            time = read_csv(
                time_file,
                index_col=False,
                delimiter=",",
                header=0,
                names=[x["column_name"] for x in csv_parsing_options],
                dtype={x["column_name"]: x["dtype"] for x in csv_parsing_options},
                on_bad_lines="skip",
                usecols=[x["column_name"] for x in csv_parsing_options if x["include"]],
            )
            file_in_range = False
            if time.shape[0] == 0:
                continue

            if start_date is not None:
                start_date_timestamp = to_datetime_utc(start_date).timestamp()
                file_in_range = not (
                    start_date_timestamp > time["time"].values[-1] + 60.0
                )

            if end_date is not None:
                end_date_timestamp = to_datetime_utc(end_date).timestamp()
                file_in_range = file_in_range and not (
                    end_date_timestamp <= time["time"].values[0]
                )

            if file_in_range:
                yield file
            else:
                continue

    else:
        for file in sorted(files):
            yield file


def load_as_dataframe(
    files_to_parse: Iterator[str],
    csv_parsing_options: List[ColumnParseParameters],
    sampling_interval=0.4,
) -> DataFrame:
    column_names = [x["column_name"] for x in csv_parsing_options]
    usecols = [x["column_name"] for x in csv_parsing_options if x["include"]]
    dtype = {x["column_name"]: x["dtype"] for x in csv_parsing_options}
    convert = [x["convert"] for x in csv_parsing_options if x["include"]]

    def process_file(file):
        df = read_csv(
            file,
            index_col=False,
            delimiter=",",
            header=0,
            names=column_names,
            dtype=dtype,
            on_bad_lines="skip",
            usecols=usecols,
        )
        df = quality_control(df)

        if "time" in df:
            df["time"] = to_datetime(df["time"], unit="s")

        for name, function in zip(usecols, convert):
            df[name] = function(df[name])

        return df

    source_files = list(files_to_parse)
    data_frames = [process_file(source_file) for source_file in source_files]

    # Fragmentation occurs for spectral data- to avoid performance issues we recreate the dataframe after the concat
    # here.
    dataframe = concat(
        data_frames, keys=source_files, names=["source files", "file index"]
    ).copy()
    dataframe.reset_index(inplace=True)
    return mark_continuous_groups(dataframe, sampling_interval)


def mark_continuous_groups(df: DataFrame, sampling_interval):
    """
    This function adds a column that has unique number for each continuous block of data (i.e. data without gaps).
    It does so by:
        - calculating the time difference
        - creating a boolean mask where the time difference is larger than a threshold
        - taking the cumulative sum over that mask- this sum mask will only increase if there is a gap between succesive
          entries - and can thus serve as a group marker.
    :param df: data frame contaning a time epoch key
    :param sampling_interval: sampling interval used
    :return:
    """

    df["group id"] = (
        df["time"].diff() > sampling_interval + timedelta64(100, "ms")
    ).cumsum()
    return df


def read_data(
    path,
    data_type,
    sampling_interval,
    start_date: datetime = None,
    end_date: datetime = None,
) -> DataFrame:
    files = files_to_parse(path, _pattern[data_type], start_date, end_date)
    format = get_format(data_type)
    dataframe = load_as_dataframe(files, format, sampling_interval)

    if start_date is not None:
        start_date = to_datetime64(start_date)
        dataframe = dataframe[dataframe["time"] >= start_date]

    if end_date is not None:
        end_date = to_datetime64(end_date)
        dataframe = dataframe[dataframe["time"] < end_date]

    return dataframe


def read_gps(
    path,
    postprocess=True,
    config: SpotterConstants = None,
    start_date: datetime = None,
    end_date: datetime = None,
) -> DataFrame:
    if config is None:
        config = spotter_constants()

    data = read_data(
        path,
        "GPS",
        sampling_interval=config["sampling_interval_gps"],
        start_date=start_date,
        end_date=end_date,
    )
    if not postprocess:
        return data

    u = data["speed over ground"].values * cos(
        (90 - data["course over ground"].values) * pi / 180
    )
    v = data["speed over ground"].values * sin(
        (90 - data["course over ground"].values) * pi / 180
    )
    w = data["w"].values
    time = data["time"].values

    latitude = data["latitude degrees"].values + data["latitude minutes"].values / 60
    longitude = data["longitude degrees"].values + data["longitude minutes"].values / 60
    z = data["z"].values

    return DataFrame(
        data={
            "time": time,
            "latitude": latitude,
            "longitude": longitude,
            "z": z,
            "u": u,
            "v": v,
            "w": w,
            "group id": data["group id"].values,
        }
    )


def read_displacement(
    path,
    postprocess=True,
    config: SpotterConstants = None,
    start_date: datetime = None,
    end_date: datetime = None,
) -> DataFrame:
    if config is None:
        config = spotter_constants()

    data = read_data(
        path,
        "FLT",
        sampling_interval=config["sampling_interval_location"],
        start_date=start_date,
        end_date=end_date,
    )
    if not postprocess:
        return data

    # Backward sos pass for phase correction
    def _process(_data: DataFrame):
        _data["x"] = sos_filter(_data["x"].values, "backward")
        _data["y"] = sos_filter(_data["y"].values, "backward")
        _data["z"] = sos_filter(_data["z"].values, "backward")
        return _data

    return apply_to_group(_process, data)


def read_location(
    path,
    postprocess=True,
    config: SpotterConstants = None,
    start_date: datetime = None,
    end_date: datetime = None,
) -> DataFrame:
    if config is None:
        config = spotter_constants()

    raw_data = read_data(
        path,
        "LOC",
        config["sampling_interval_location"],
        start_date=start_date,
        end_date=end_date,
    )
    if not postprocess:
        return raw_data

    dataframe = DataFrame()
    dataframe["time"] = raw_data["time"]
    dataframe["latitude"] = (
        raw_data["latitude degrees"] + raw_data["latitude minutes"] / 60
    )
    dataframe["longitude"] = (
        raw_data["longitude degrees"] + raw_data["longitude minutes"] / 60
    )
    dataframe["group id"] = raw_data["group id"]
    return dataframe


def read_raw_spectra(
    path,
    postprocess=True,
    config: SpotterConstants = None,
    start_date: datetime = None,
    end_date: datetime = None,
) -> DataFrame:
    if config is None:
        config = spotter_constants()
    if not postprocess:
        return read_data(
            path,
            "SPC",
            config["sampling_interval_spectra"],
            start_date=start_date,
            end_date=end_date,
        )

    if config is None:
        config = spotter_constants()

    column_names = spectral_column_names(groupedby="kind", config=config)

    dataframes = []
    raw_data = read_data(path, "SPC", config["sampling_interval_spectra"])

    data_types = []
    for data_type, freq_column_names in column_names:
        df = raw_data[["time"] + freq_column_names]
        rename_mapping = {name: index for index, name in enumerate(freq_column_names)}
        df = df.rename(rename_mapping, axis=1)
        data_types.append(data_type)
        dataframes.append(df)

    data = concat(dataframes, keys=data_types, names=["kind", "source_index"])
    data.reset_index(inplace=True)
    data.drop(columns="source_index", inplace=True)
    return data


def read_spectra(
    path,
    depth=inf,
    config: SpotterConstants = None,
    start_date: datetime = None,
    end_date: datetime = None,
) -> FrequencySpectrum:
    if config is None:
        config = spotter_constants()

    data = read_raw_spectra(path, start_date=start_date, end_date=end_date)
    sampling_interval = config["sampling_interval_gps"] / timedelta64(1, "s")
    df = 1 / (config["number_of_samples"] * sampling_interval)
    frequencies = (
        linspace(
            0,
            config["number_of_frequencies"],
            config["number_of_frequencies"],
            endpoint=False,
        )
        * df
    )

    spectral_values = data[list(range(0, config["number_of_frequencies"]))].values
    time = data["time"].values
    time = time[data["kind"] == "Szz_re"]

    Szz = spectral_values[data["kind"] == "Szz_re", :]
    Sxx = spectral_values[data["kind"] == "Sxx_re", :]
    Syy = spectral_values[data["kind"] == "Syy_re", :]
    Cxy = spectral_values[data["kind"] == "Sxy_re", :]
    Qzx = spectral_values[data["kind"] == "Szx_im", :]
    Qzy = spectral_values[data["kind"] == "Szy_im", :]

    with errstate(invalid="ignore", divide="ignore"):
        # Supress divide by 0; silently produce NaN
        a1 = Qzx / sqrt(Szz * (Sxx + Syy))
        a2 = (Sxx - Syy) / (Sxx + Syy)
        b1 = Qzy / sqrt(Szz * (Sxx + Syy))
        b2 = 2.0 * Cxy / (Sxx + Syy)

    location = read_location(path, postprocess=True, config=config)
    latitude = interp(
        datetime64_to_timestamp(time),
        datetime64_to_timestamp(location["time"].values),
        location["latitude"].values,
    )
    longitude = interp(
        datetime64_to_timestamp(time),
        datetime64_to_timestamp(location["time"].values),
        location["longitude"].values,
    )

    depth = full_like(time, depth)

    dataset = Dataset(
        data_vars={
            "variance_density": (["time", "frequency"], Szz),
            "a1": (["time", "frequency"], a1),
            "b1": (["time", "frequency"], b1),
            "a2": (["time", "frequency"], a2),
            "b2": (["time", "frequency"], b2),
            "depth": (["time"], depth),
            "latitude": (["time"], latitude),
            "longitude": (["time"], longitude),
        },
        coords={"time": to_datetime(time), "frequency": frequencies},
    )
    return FrequencySpectrum(dataset)


def quality_control(dataframe: DataFrame) -> DataFrame:
    negative_time = dataframe["time"].diff() < 0.0
    dataframe.loc[negative_time, "time"] = nan
    dataframe.dropna(inplace=True)
    return dataframe



from roguewave.timeseries_analysis.filtering import (
    sos_filter,
    spike_filter,
    exponential_delta_filter,
    exponential_filter,
    cumulative_filter,
)
from roguewave.timeseries_analysis.time_integration import integrate
from numpy.typing import NDArray
from numpy import issubdtype, datetime64
from roguewave.tools.time import datetime64_to_timestamp
from typing import Union, List, Callable
from scipy.signal import butter

sos = butter(4, 0.04, btype="high", output="sos", fs=2.5)


def pipeline(time: NDArray, signal: NDArray, stages=None) -> NDArray:
    if stages is None:
        stages = [
            ("spike", None),
            ("exponential_delta", None),
            ("integrate", None),
            ("sos_filtfilt", None),
        ]

    stages = _create_pipeline(stages)

    if issubdtype(time.dtype, datetime64):
        time = datetime64_to_timestamp(time)

    for stage in stages:
        signal = stage(time, signal)

    return signal


STAGE_SIGNATURE = Callable[[NDArray, NDArray], NDArray]


def create_stage(stage: str, stage_opt=None) -> STAGE_SIGNATURE:
    if stage == "sos_forward":
        return lambda time, signal: sos_filter(signal, "forward", stage_opt)
    elif stage == "sos_backward":
        return lambda time, signal: sos_filter(signal, "backward", stage_opt)
    elif stage == "sos_filtfilt":
        return lambda time, signal: sos_filter(signal, "filtfilt", stage_opt)
    elif stage == "integrate":
        return lambda time, signal: integrate(time, signal)
    elif stage == "spike":
        return lambda time, signal: spike_filter(time, signal)
    elif stage == "exponential_delta":
        return lambda time, signal: exponential_delta_filter(time, signal)
    elif stage == "cumulative":
        return lambda time, signal: cumulative_filter(time, signal)
    elif stage == "exponential":
        return lambda time, signal: exponential_filter(time, signal)
    else:
        raise Exception(f"unknown filter: {stage}")


def _create_pipeline(
    stages: List[Union[str, STAGE_SIGNATURE]]
) -> List[STAGE_SIGNATURE]:
    functions = []
    for stage in stages:
        if isinstance(stage, tuple):
            functions.append(create_stage(stage=stage[0], stage_opt=stage[1]))
        elif isinstance(stage, str):
            functions.append(create_stage(stage=stage, stage_opt=None))
        else:
            functions.append(stage)
    return functions



from .read_csv_data import (
    read_spectra,
    read_displacement,
    read_gps,
    read_raw_spectra,
    read_location,
)

from .analysis import (
    spectra_from_raw_gps,
    spectra_from_displacement,
    displacement_from_gps_doppler_velocities,
    displacement_from_gps_positions,
)



"""
Contents: IO

Copyright (C) 2022
Sofar Ocean Technologies

Authors: Pieter Bart Smit
======================

Routines that can be used to save and load data.

Functions:

How To Use This Module
======================
(See the individual functions for details.)

"""
from roguewave.wavespectra import (
    FrequencySpectrum,
    FrequencyDirectionSpectrum,
    create_1d_spectrum,
)
from pandas import DataFrame, read_json
from typing import Union, Dict, List
from datetime import datetime
import gzip
import json
import os
import numpy
import base64
from io import BytesIO
from xarray import Dataset, open_dataset
from roguewave import to_datetime_utc


_UNION = Union[
    FrequencySpectrum,
    FrequencyDirectionSpectrum,
    List[FrequencySpectrum],
    Dict[str, List[FrequencyDirectionSpectrum]],
    Dict[str, List[FrequencySpectrum]],
    List[List[FrequencyDirectionSpectrum]],
    List[List[DataFrame]],
    Dict[int, List[DataFrame]],
    Dict[str, DataFrame],
]


def _b64_encode_numpy(val: numpy.ndarray) -> dict:
    if val.dtype == numpy.float64:
        val = val.astype(dtype=numpy.float32)

    if val.dtype == numpy.int64:
        val = val.astype(dtype=numpy.int32)

    dtype = val.dtype.str
    shape = val.shape
    data = base64.b64encode(val)

    return {
        "__class__": "ndarray",
        "data": {"dtype": dtype, "shape": shape, "data": data.decode("utf8")},
    }


def _b64_decode_numpy(data):
    decoded_bytes = base64.b64decode(bytes(data["data"], encoding="utf-8"))
    dtype = numpy.dtype(data["dtype"])
    array = numpy.frombuffer(decoded_bytes, dtype=dtype)
    return numpy.reshape(array, data["shape"])


def _b64_encode_dataset(val: Dataset, name) -> dict:
    net_cdf = val.to_netcdf(engine="scipy")
    data = base64.b64encode(net_cdf)
    return {"__class__": name, "data": {"data": data.decode("utf8")}}


def _b64_decode_dataset(data):
    decoded_bytes = base64.b64decode(bytes(data["data"], encoding="utf-8"))
    with BytesIO(decoded_bytes) as fp:
        dataset = open_dataset(fp, engine="scipy")
    return dataset


class NumpyEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, numpy.ndarray):
            return _b64_encode_numpy(obj)
        elif isinstance(obj, numpy.int32):
            return int(obj)
        elif isinstance(obj, numpy.float32):
            return float(obj)
        elif isinstance(obj, numpy.int64):
            return int(obj)
        elif isinstance(obj, numpy.float64):
            return float(obj)
        elif isinstance(obj, Dataset):
            return _b64_encode_dataset(obj, "Dataset")
        elif isinstance(obj, FrequencySpectrum):
            return _b64_encode_dataset(obj.dataset, "FrequencySpectrum")
        elif isinstance(obj, FrequencyDirectionSpectrum):
            return _b64_encode_dataset(obj.dataset, "FrequencyDirectionSpectrum")
        elif isinstance(obj, datetime):
            return {"__class__": "datetime", "data": datetime.isoformat(obj)}
        elif isinstance(obj, DataFrame):
            return {"__class__": "DataFrame", "data": obj.to_json(date_unit="s")}
        else:
            return json.JSONEncoder.default(self, obj)


def object_hook(dictionary: dict):
    if "__class__" in dictionary:
        if dictionary["__class__"] == "DataFrame":
            #
            df = read_json(dictionary["data"])
            if "timestamp" in df:
                df["timestamp"] = df["timestamp"].apply(lambda x: x.tz_localize("utc"))
            elif "time" in df:
                time = to_datetime_utc(df["time"].values)
                # df["time"] = df["time"].apply(lambda x: x.tz_localize("utc"))
                df["time"] = time
            elif "valid_time" in df:
                df["valid_time"] = df["valid_time"].apply(
                    lambda x: x.tz_localize("utc")
                )
            elif "init_time" in df:
                df["valid_time"] = df["valid_time"].apply(
                    lambda x: x.tz_localize("utc")
                )
            else:
                df.index = [x.tz_localize("utc") for x in df.index]
            return df
        elif dictionary["__class__"] == "WaveSpectrum1D":
            data = dictionary["data"]
            return create_1d_spectrum(
                frequency=data["frequency"],
                variance_density=data["varianceDensity"],
                time=data["timestamp"],
                latitude=data["latitude"],
                longitude=data["longitude"],
                a1=data["a1"],
                b1=data["b1"],
                a2=data["a2"],
                b2=data["b2"],
            )
        elif dictionary["__class__"] == "Dataset":
            return _b64_decode_dataset(dictionary["data"])
        elif dictionary["__class__"] == "FrequencySpectrum":
            return FrequencySpectrum(_b64_decode_dataset(dictionary["data"]))
        elif dictionary["__class__"] == "FrequencyDirectionSpectrum":
            return FrequencyDirectionSpectrum(_b64_decode_dataset(dictionary["data"]))
        elif dictionary["__class__"] == "ndarray":
            return _b64_decode_numpy(dictionary["data"])
        elif dictionary["__class__"] == "datetime":
            return datetime.fromisoformat(dictionary["data"])
    else:
        return dictionary


def load(filename: str) -> _UNION:
    """
    Load spectral data as saved by "save_spectrum" from the given file and
    return a (nested) object. The precise format of the output depends on what
    was saved.

    :param filename: path to file to load.
    :return:
        - Data in the same form it was saved.
    """

    with gzip.open(filename, "rb") as file:
        data = file.read().decode("utf-8")
        return json.loads(data, object_hook=object_hook)


def save(
    _input: _UNION, filename: str, format="json", separate_spotters_into_files=False
):
    """
    Save spectral data, possible in nested form as returned by the spectral
    partition/reconstruction/etc. functions. Data is saved in JSON form
    compressed with gzip.

    :param _input:
        - Data containing python primitives (dict/list) or any of the RogueWave
          classes.

    :param filename: path to save data.
    :return: None
    """

    def write(filename, _input, format):
        if format == "json":
            with gzip.open(filename, "wt") as file:
                json.dump(_input, file, cls=NumpyEncoder)
                # file.write(data.encode('utf-8'))
        else:
            raise Exception("Unknown output format")

    if separate_spotters_into_files:
        os.makedirs(filename, exist_ok=True)
        for key in _input:
            name = os.path.join(filename, key)
            write(name, _input[key], format)

    else:
        write(filename, _input, format)



from numba import njit
from numba.typed import Dict
from numba.core import types
from numpy import empty_like, abs, nan, isfinite, flip, linspace
from numpy.typing import NDArray
from typing import Literal
from numpy import interp
from scipy.signal import sosfilt, sosfiltfilt, butter


@njit(cache=True)
def exponential_filter(
    time_seconds, signal, sampling_frequency=2.5, options: dict = None
):
    """
    Exponential filter that operates on the differences between succesive values.

    :param time_seconds:
    :param signal:
    :param options:
    :return:
    """

    if options is None:
        options = Dict.empty(key_type=types.unicode_type, value_type=types.float64)

    smoothing_factor = options.get("smoothing_factor", 0.004)
    maximum_gap_size = options.get("maximum_gap_size", 3)
    sampling_interval = 1 / sampling_frequency

    # Initialize empty array
    filtered_signal = empty_like(signal)

    # Initialize start values
    exponential_mean = 0
    filtered_signal[0] = 0

    for ii in range(1, len(signal)):
        if (
            time_seconds[ii] - time_seconds[ii - 1]
            > maximum_gap_size * sampling_interval
        ):
            # Restart
            exponential_mean = 0.0
            filtered_signal[ii] = 0.0

        exponential_mean = (
            exponential_mean * (1 - smoothing_factor) + signal[ii] * smoothing_factor
        )
        filtered_signal[ii] = signal[ii] - exponential_mean

    return filtered_signal


@njit(cache=True)
def exponential_delta_filter(
    time_seconds, signal, sampling_frequency=2.5, options: dict = None
):
    """
    Exponential filter that operates on the differences between succesive values.

    :param time_seconds:
    :param signal:
    :param options:
    :return:
    """

    if options is None:
        options = Dict.empty(key_type=types.unicode_type, value_type=types.float64)

    smoothing_factor = options.get("smoothing_factor", 0.004)
    maximum_gap_size = options.get("maximum_gap_size", 3)
    sampling_interval = 1 / sampling_frequency

    # Initialize empty array
    filtered_signal = empty_like(signal)

    # Initialize start values
    exponential_mean = 0
    filtered_signal[0] = 0

    for ii in range(1, len(signal)):
        signal_delta = signal[ii] - signal[ii - 1]

        if (
            time_seconds[ii] - time_seconds[ii - 1]
            > maximum_gap_size * sampling_interval
        ):
            # Restart
            exponential_mean = 0.0
            filtered_signal[ii] = 0.0

        exponential_mean = (
            exponential_mean * (1 - smoothing_factor) + signal_delta * smoothing_factor
        )
        filtered_signal[ii] = filtered_signal[ii - 1] + (
            signal_delta - exponential_mean
        )

    return filtered_signal


@njit(cache=True)
def spike_filter(time: NDArray, signal: NDArray, options: dict = None) -> NDArray:
    if options is None:
        options = Dict.empty(key_type=types.unicode_type, value_type=types.float64)

    rejection_threshold = options.get("rejection_threshold", 9.81)

    filtered_signal = empty_like(signal)
    filtered_signal[0] = signal[0]

    prev_signal = signal[0]
    prev_time = time[0]

    for ii in range(1, len(signal)):
        delta_signal = signal[ii] - prev_signal
        delta_time = time[ii] - prev_time

        if abs(delta_signal / delta_time) > rejection_threshold:
            filtered_signal[ii] = nan
            continue

        filtered_signal[ii] = prev_signal + delta_signal
        prev_signal = filtered_signal[ii]
        prev_time = time[ii]
    return nan_interpolate(time, filtered_signal)


@njit(cache=True)
def cumulative_filter(time: NDArray, signal: NDArray, options: dict = None) -> NDArray:
    if options is None:
        options = Dict.empty(key_type=types.unicode_type, value_type=types.float64)

    smoothing_factor = options.get("smoothing_factor", 0.01)

    filtered_signal = empty_like(signal)
    filtered_signal[0] = signal[0]
    cumchange = 0.0

    lf = 0.0
    istart = 0
    for ii in range(1, len(signal)):
        delta_signal = signal[ii] - signal[ii - 1]

        lf = lf * (1 - smoothing_factor) + delta_signal * smoothing_factor
        delta = delta_signal - lf

        if cumchange * (cumchange + delta) < 0:
            istart = ii
        else:
            if ii - istart > 23:
                filtered_signal[istart:ii] = filtered_signal[
                    istart:ii
                ] - cumchange * linspace(0, 1, ii - istart)
                istart = ii
                cumchange = 0.0

        cumchange += delta
        filtered_signal[ii] = filtered_signal[ii - 1] + delta

    return nan_interpolate(time, filtered_signal)


def sos_filter(
    signal: NDArray, direction: Literal["backward", "forward", "filtfilt"], sos=None
) -> NDArray:
    #
    # Apply forward/backward/filtfilt sos filter
    #
    # Get SOS coefficients
    if sos is None:
        sos = butter(4, 0.033, btype="high", output="sos", fs=2.5)
    #
    if direction == "backward":
        return flip(sosfilt(sos, flip(signal)))

    elif direction == "forward":
        return sosfilt(sos, signal)

    elif direction == "filtfilt":
        return sosfiltfilt(sos, signal)

    else:
        raise ValueError(f"Unknown direction {direction}")


@njit(cache=True)
def nan_interpolate(time: NDArray, signal: NDArray) -> NDArray:
    mask = isfinite(signal)
    return interp(time, time[mask], signal[mask])



from pandas import read_csv, concat, DataFrame, Timestamp
from roguewave import to_datetime_utc
from numpy import pi, cos, sin, int64, int32, float32, float64
from typing import Iterator, TypeVar
from glob import glob
import os

_T = TypeVar("_T")


def do_nothing(x: _T) -> _T:
    return x


GPS_HEADER = {
    "millis": {"dtype": float32, "include": False, "convert": do_nothing},
    "GPS_Epoch_Time(s)": {"dtype": float64, "include": True, "convert": do_nothing},
    "lat(deg)": {"dtype": float32, "include": True, "convert": do_nothing},
    "lat(min*1e5)": {"dtype": float32, "include": True, "convert": lambda x: x / 60e5},
    "long(deg)": {"dtype": float32, "include": True, "convert": do_nothing},
    "long(min*1e5)": {"dtype": float32, "include": True, "convert": lambda x: x / 60e5},
    "el(m)": {"dtype": float32, "include": True, "convert": do_nothing},
    "SOG(mm/s)": {"dtype": float32, "include": True, "convert": lambda x: x / 1000},
    "COG(deg*1000)": {"dtype": float32, "include": True, "convert": lambda x: x / 1000},
    "vert_vel(mm/s)": {
        "dtype": float32,
        "include": True,
        "convert": lambda x: x / 1000,
    },
}

DISP_CSV_PARSING = {
    "millis": {"dtype": float32, "include": False, "convert": do_nothing},
    "time": {"dtype": float64, "include": True, "convert": do_nothing},
    "x": {"dtype": float32, "include": True, "convert": lambda x: x / 1000},
    "y": {"dtype": float32, "include": True, "convert": lambda x: x / 1000},
    "z": {"dtype": float32, "include": True, "convert": lambda x: x / 1000},
    "init": {"dtype": str, "include": False, "convert": do_nothing},
}


def files_to_parse(path: str, pattern: str) -> Iterator[str]:
    files = glob(os.path.join(path, pattern))
    for file in sorted(files):
        yield file


def load_as_dataframe(
    files_to_parse: Iterator[str],
    csv_parsing_options: dict,
    postprocess_dataframe=lambda x: x,
) -> DataFrame:
    names = list(csv_parsing_options.keys())
    columns = [key for key in names if csv_parsing_options[key]["include"]]
    dtype = {key: csv_parsing_options[key]["dtype"] for key in names}

    def process_file(file):
        df = read_csv(
            file,
            delimiter=",",
            header=0,
            names=names,
            dtype=dtype,
            on_bad_lines="skip",
            usecols=columns,
        )
        df.dropna(inplace=True)

        for name in columns:
            df[name] = csv_parsing_options[name]["convert"](df[name])

        return postprocess_dataframe(df)

    source_files = list(files_to_parse)
    data_frames = [process_file(source_file) for source_file in source_files]
    return mark_continuous_groups(concat(data_frames, keys=source_files))


def mark_continuous_groups(df: DataFrame, sampling_interval=0.4):
    df["continuous_group_marker"] = (
        df["time"].diff() > sampling_interval + 0.01
    ).cumsum()
    return df


def load_gps(path: str) -> DataFrame:
    def process(raw_dataframe: DataFrame) -> DataFrame:
        dataframe = DataFrame()
        dataframe["time"] = raw_dataframe["GPS_Epoch_Time(s)"].values
        dataframe["latitude"] = (
            raw_dataframe["lat(deg)"].values
            + raw_dataframe["lat(min*1e5)"].values / 60e5
        )
        dataframe["longitude"] = (
            raw_dataframe["long(deg)"].values
            + raw_dataframe["long(min*1e5)"].values / 60e5
        )
        dataframe["z"] = raw_dataframe["el(m)"].values

        angle = (90 - raw_dataframe["COG(deg*1000)"].values / 1000) * pi / 180
        dataframe["u"] = cos(angle) * raw_dataframe["SOG(mm/s)"].values
        dataframe["v"] = sin(angle) * raw_dataframe["SOG(mm/s)"].values
        dataframe["w"] = raw_dataframe["vert_vel(mm/s)"].values
        return dataframe

    return load_as_dataframe(
        files_to_parse(path, "????_GPS.csv"),
        csv_parsing_options=GPS_HEADER,
        postprocess_dataframe=process,
    )


def load_displacement(path):
    # millis, GPS_Epoch_Time(s), outx(mm), outy(mm), outz(mm)
    return load_as_dataframe(
        files_to_parse(path, "????_FLT.csv"),
        csv_parsing_options=DISP_CSV_PARSING,
        postprocess_dataframe=do_nothing,
    )


if __name__ == "__main__":
    path = "/Users/pietersmit/Downloads/Sunflower13/log"
    df = load_gps(path)
    print(df.shape)



from numpy import float32, float64, int32, timedelta64
from typing import TypeVar, List, TypedDict, Callable
from ._spotter_constants import SpotterConstants, spotter_constants

_T = TypeVar("_T")


def do_nothing(x: _T) -> _T:
    return x


class ColumnParseParameters(TypedDict):
    header_name: str
    column_name: str
    dtype: type
    include: bool
    convert: Callable[[_T], _T]


_formats = {
    "GPS": [
        ColumnParseParameters(
            header_name="millis",
            column_name="milliseconds",
            dtype=float32,
            include=False,
            convert=do_nothing,
        ),
        ColumnParseParameters(
            header_name="GPS_Epoch_Time(s)",
            column_name="time",
            dtype=float64,
            include=True,
            convert=do_nothing,
        ),
        ColumnParseParameters(
            header_name="lat(deg)",
            column_name="latitude degrees",
            dtype=float64,
            include=True,
            convert=do_nothing,
        ),
        ColumnParseParameters(
            header_name="lat(min*1e5)",
            column_name="latitude minutes",
            dtype=float64,
            include=True,
            convert=lambda x: x / 1e5,
        ),
        ColumnParseParameters(
            header_name="lon(deg)",
            column_name="longitude degrees",
            dtype=float64,
            include=True,
            convert=do_nothing,
        ),
        ColumnParseParameters(
            header_name="long(min*1e5)",
            column_name="longitude minutes",
            dtype=float64,
            include=True,
            convert=lambda x: x / 1e5,
        ),
        ColumnParseParameters(
            header_name="el(m)",
            column_name="z",
            dtype=float32,
            include=True,
            convert=do_nothing,
        ),
        ColumnParseParameters(
            header_name="SOG(mm/s)",
            column_name="speed over ground",
            dtype=float32,
            include=True,
            convert=lambda x: x / 1000,
        ),
        ColumnParseParameters(
            header_name="COG(deg*1000)",
            column_name="course over ground",
            dtype=float32,
            include=True,
            convert=lambda x: x / 1000,
        ),
        ColumnParseParameters(
            header_name="vert_vel(mm/s)",
            column_name="w",
            dtype=float32,
            include=True,
            convert=lambda x: x / 1000,
        ),
    ],
    "FLT": [
        ColumnParseParameters(
            header_name="millis",
            column_name="milliseconds",
            dtype=float32,
            include=False,
            convert=do_nothing,
        ),
        ColumnParseParameters(
            header_name="GPS_Epoch_Time(s)",
            column_name="time",
            dtype=float64,
            include=True,
            convert=do_nothing,
        ),
        ColumnParseParameters(
            header_name="outx(mm)",
            column_name="x",
            dtype=float32,
            include=True,
            convert=lambda x: x / 1000,
        ),
        ColumnParseParameters(
            header_name="outy(mm)",
            column_name="y",
            dtype=float32,
            include=True,
            convert=lambda x: x / 1000,
        ),
        ColumnParseParameters(
            header_name="outz(mm)",
            column_name="z",
            dtype=float32,
            include=True,
            convert=lambda x: x / 1000,
        ),
        ColumnParseParameters(
            header_name="",
            column_name="init flag",
            dtype=str,
            include=False,
            convert=do_nothing,
        ),
    ],
    "LOC": [
        ColumnParseParameters(
            header_name="GPS_Epoch_Time(s)",
            column_name="time",
            dtype=float64,
            include=True,
            convert=do_nothing,
        ),
        ColumnParseParameters(
            header_name="lat(deg)",
            column_name="latitude degrees",
            dtype=float32,
            include=True,
            convert=do_nothing,
        ),
        ColumnParseParameters(
            header_name="at(min*1e5)",
            column_name="latitude minutes",
            dtype=float32,
            include=True,
            convert=lambda x: x / 1e5,
        ),
        ColumnParseParameters(
            header_name="lon(deg)",
            column_name="longitude degrees",
            dtype=float32,
            include=True,
            convert=do_nothing,
        ),
        ColumnParseParameters(
            header_name="long(min*1e5)",
            column_name="longitude minutes",
            dtype=float32,
            include=True,
            convert=lambda x: x / 1e5,
        ),
    ],
    "TIME": [
        ColumnParseParameters(
            header_name="GPS_Epoch_Time(s)",
            column_name="time",
            dtype=float64,
            include=True,
            convert=do_nothing,
        ),
        ColumnParseParameters(
            header_name="lat(deg)",
            column_name="latitude degrees",
            dtype=float32,
            include=False,
            convert=do_nothing,
        ),
        ColumnParseParameters(
            header_name="at(min*1e5)",
            column_name="latitude minutes",
            dtype=float32,
            include=False,
            convert=lambda x: x / 1e5,
        ),
        ColumnParseParameters(
            header_name="lon(deg)",
            column_name="longitude degrees",
            dtype=float32,
            include=False,
            convert=do_nothing,
        ),
        ColumnParseParameters(
            header_name="long(min*1e5)",
            column_name="longitude minutes",
            dtype=float32,
            include=False,
            convert=lambda x: x / 1e5,
        ),
    ],
    "SPC": [
        ColumnParseParameters(
            header_name="type",
            column_name="type",
            dtype=str,
            include=False,
            convert=do_nothing,
        ),
        ColumnParseParameters(
            header_name="millis",
            column_name="milliseconds",
            dtype=float32,
            include=False,
            convert=do_nothing,
        ),
        ColumnParseParameters(
            header_name="t0_GPS_Epoch_Time(s)",
            column_name="time",
            dtype=float64,
            include=True,
            convert=do_nothing,
        ),
        ColumnParseParameters(
            header_name="tN_GPS_Epoch_Time(s)",
            column_name="tN time",
            dtype=float64,
            include=True,
            convert=do_nothing,
        ),
        ColumnParseParameters(
            header_name="ens_count",
            column_name="ensemble count",
            dtype=int32,
            include=True,
            convert=do_nothing,
        ),
    ],
}

_spectral_column_names = {
    "default": [
        "Sxx_re",
        "Syy_re",
        "Szz_re",
        "Sxy_re",
        "Szx_re",
        "Szy_re",
        "Sxx_im",
        "Syy_im",
        "Szz_im",
        "Sxy_im",
        "Szx_im",
        "Szy_im",
    ]
}


def spectral_column_names(groupedby="frequency", config: SpotterConstants = None):
    if config is None:
        config = spotter_constants()

    if config["n_channel"] is False:
        names = _spectral_column_names["default"]
    else:
        raise ValueError("not implemented yet")

    columns = []
    if groupedby == "frequency":
        for index in range(0, config["number_of_frequencies"]):
            columns.append((index, [name + f"_{index}" for name in names]))
        return columns
    if groupedby == "flat":
        for index in range(0, config["number_of_frequencies"]):
            columns += [name + f"_{index}" for name in names]
        return columns
    elif groupedby == "kind":
        for name in names:
            columns.append(
                (
                    name,
                    [
                        name + f"_{index}"
                        for index in range(0, config["number_of_frequencies"])
                    ],
                )
            )
        return columns
    else:
        raise ValueError("Unknown grouping")


def create_spectral_format(config: SpotterConstants) -> List[ColumnParseParameters]:
    if config is None:
        config = spotter_constants()

    _format = _formats["SPC"]
    sampling_interval = config["sampling_interval_gps"] / timedelta64(1, "s")

    sampling_frequency = 1 / (config["number_of_samples"] * sampling_interval)

    def _create(name) -> ColumnParseParameters:
        return ColumnParseParameters(
            header_name=name,
            column_name=name,
            dtype=float32,
            include=True,
            convert=lambda x: x / (1000000.0 * sampling_frequency),
        )

    column_names = spectral_column_names("flat", config)
    return _format + [_create(name) for name in column_names]


def get_format(
    file_type, config: SpotterConstants = None
) -> List[ColumnParseParameters]:
    if file_type == "SPC":
        return create_spectral_format(config)
    else:
        return _formats[file_type]



# Author: Pieter Bart Smit

"""
A set of spectral analysis routines specifically designed to work with data
from Sofar Spotter ocean wave buoys.

Classes:

- `SpectralAnalysisConfig`, configuration object

Functions:

- `segment_timeseries`, split timeseries into non-overlapping segments
- `calculate_moments`, calculate directional moments from (co)variances
- `spectral_analysis`, calculate wave spectrum for specific segment
- `generate_spectra`, calculate wave spectra for all segments
- `window_power_correction_factor`, correct for power loss due to windowing

How To Use This Module
======================
(See the individual classes, methods, and attributes for details.)

1. Import it: ``import welch`` or ``from welch import ...``.
s
"""

import numpy
from numpy.fft import fft
from typing import List
from roguewave import FrequencySpectrum
from numpy.typing import NDArray
from typing import Tuple
from scipy.signal.windows import get_window
from xarray import Dataset, DataArray
from numba import njit, objmode


@njit(cache=True)
def segment_timeseries(epoch_time, segment_length_seconds) -> List[tuple]:
    """
    This function segments the time series into segments of the given length.
    It returns a list of tuples containing the indices of the start and end
    point of each segment. Segmens are open ended [t_start,t_end>.

    If the segment length does not fit a whole number of times into the vector
    the partial remaining segment is discarded.

    :param epoch_time: time vector (seconds, unix epoch)
    :param segment_length_seconds: lenght of segment in segments
    :return:
    """

    # Initialize the start time and the expected next segment start
    istart = 0
    t_start = epoch_time[0]
    t_next = t_start + segment_length_seconds

    segments = []
    # loop over all times
    for index, time in enumerate(epoch_time):
        # if the current time is larger or equal to the next segment start,
        # create a new segment
        if time > t_next:
            # Add segment to list
            segments.append((istart, index))

            # setup the new start time/index and next expected segment start
            istart = index
            t_start = t_next
            t_next = t_start + segment_length_seconds

    return segments


@njit(cache=True)
def calculate_co_spectra(
    signals,
    window: numpy.ndarray,
    sampling_frequency,
    window_overlap_fraction=0.5,
    spectral_window=None,
) -> NDArray:
    overlap = int(len(window) * window_overlap_fraction)
    nyquist_index = len(window) // 2

    nsig = len(signals)
    nt = len(signals[0])
    output = numpy.zeros((nsig, nsig, nyquist_index), dtype="complex_")
    ffts = numpy.zeros((nsig, nyquist_index), dtype="complex_")

    zero_mean_signals = numpy.zeros((nsig, nt))
    for index, signal in enumerate(signals):
        zero_mean_signals[index, :] = signal - numpy.mean(signal)

    ii = 0
    istart = 0
    while True:
        # Advance the counter
        iend = istart + len(window)
        if iend > nt:
            break

        for index in range(nsig):
            with objmode():
                # FFT not supported in Numba- yet
                ffts[index, :] = fft(zero_mean_signals[index, istart:iend] * window)[
                    0:nyquist_index
                ]

        sigs = numpy.zeros((nsig, nyquist_index))
        for index in range(nsig):
            sigs[index, :] = numpy.real(
                ffts[index, :] * numpy.conjugate(ffts[index, :])
            )

        ii += 1
        for mm in range(0, nsig):
            for nn in range(mm, nsig):
                output[mm, nn, :] += 2 * ffts[mm, :] * numpy.conjugate(ffts[nn, :])

        istart = istart + len(window) - overlap

    frequency_step = sampling_frequency / len(window)
    factor = (
        window_power_correction_factor(window) ** 2
        / ii
        / frequency_step
        / len(window) ** 2
    )

    for mm in range(0, nsig):
        for nn in range(mm, nsig):

            if spectral_window is not None:
                scaling = numpy.sum(output[mm, nn, :])

                if not (numpy.abs(scaling) == 0.0):
                    n = len(spectral_window) // 2
                    smoothed = numpy.convolve(output[mm, nn, :], spectral_window)

                    output[mm, nn, :] = (
                        scaling * smoothed[n:-n] / numpy.sum(smoothed[n:-n])
                    )

            output[mm, nn, :] = output[mm, nn, :] * factor

            if mm != nn:
                output[nn, mm] = output[mm, nn]

    return output


@njit(cache=True)
def welch(
    epoch_time: numpy.ndarray,
    signals,
    window: numpy.ndarray,
    sampling_frequency: float,
    segment_length_seconds=1800,
    window_overlap_fraction=0.5,
    spectral_window=None,
) -> Tuple[NDArray, NDArray, NDArray]:
    segments = segment_timeseries(epoch_time, segment_length_seconds)

    number_of_spectra = len(segments)
    number_of_frequencies = len(window) // 2

    nsig = len(signals)
    co_spectra = numpy.empty(
        (number_of_spectra, nsig, nsig, number_of_frequencies), dtype="complex_"
    )
    spectral_time = numpy.empty(number_of_spectra)

    for index, segment in enumerate(segments):
        istart = segment[0]
        iend = segment[1]
        spectral_time[index] = (epoch_time[istart] + epoch_time[iend]) / 2

        segment_signal = [sig[istart:iend] for sig in signals]
        co_spectra[index, :, :, :] = calculate_co_spectra(
            segment_signal,
            window,
            sampling_frequency,
            window_overlap_fraction,
            spectral_window,
        )

    df = sampling_frequency / len(window)
    frequencies = (
        numpy.linspace(numpy.int64(0), len(window) // 2 - 1, len(window) // 2) * df
    )

    return spectral_time, frequencies, co_spectra


@njit(cache=True)
def window_power_correction_factor(window: numpy.ndarray) -> float:
    """
    Calculate the power correction factor that corrects for the loss in power
    due to the aplication of the windowing function.

    Basically for a DC signal we want the windowed mean power to be equal to the non-windowed power

    mean (   (1 * ScaledWindow) **2 )  = mean( 1 )

    whereas the raw window gives

    Power = mean (   (Window) **2 )

    Hence- ScaledWindow = Window * sqrt( 1 / mean(window**2) )

    :param window: Window as numpy ndarray
    :return:
    """
    return 1 / numpy.sqrt(numpy.mean(window**2))


def extract_moments(co_spectra, index_x, index_y, index_z):
    szz = numpy.real(co_spectra[:, index_z, index_z, :])
    syy = numpy.real(co_spectra[:, index_y, index_y, :])
    sxx = numpy.real(co_spectra[:, index_x, index_x, :])
    cxy = numpy.real(co_spectra[:, index_x, index_y, :])
    qzx = numpy.imag(co_spectra[:, index_z, index_x, :])
    qzy = numpy.imag(co_spectra[:, index_z, index_y, :])

    a1 = qzx / numpy.sqrt(szz * (sxx + syy))
    b1 = qzy / numpy.sqrt(szz * (sxx + syy))
    a2 = (sxx - syy) / (sxx + syy)
    b2 = 2 * cxy / (sxx + syy)
    return szz, a1, b1, a2, b2


def estimate_co_spectra(
    epoch_time,
    signals,
    window=None,
    segment_length_seconds=1800,
    window_overlap_fraction=0.5,
    sampling_frequency=2.5,
    spectral_window=None,
) -> Tuple[NDArray, NDArray, NDArray]:
    if window is None:
        window = get_window("hann", 256)

    return welch(
        epoch_time,
        signals,
        window,
        sampling_frequency,
        segment_length_seconds,
        window_overlap_fraction,
        spectral_window,
    )


def estimate_frequency_spectrum(
    epoch_time,
    x,
    y,
    z,
    window=None,
    segment_length_seconds=1800,
    window_overlap_fraction=0.5,
    sampling_frequency=2.5,
    depth: DataArray = None,
    latitude: DataArray = None,
    longitude: DataArray = None,
    spectral_window=None,
    **kwargs
) -> FrequencySpectrum:
    """

    :param epoch_time: epoch time in seconds
    :param x: East(positive)/west displacement
    :param y: North(positive)/south displacement
    :param z: vertical (up positive) displacement
    :param window: window function to apply. If None are given a 256 point hann window is applied (default on Spotter),
                   which corresponds to 102.4 second length window at a sampling frequency of 2.5 Hz
                   (default on Spotter)
    :param segment_length_seconds:  Segment length for the spectral analysis. If none is given a 1800
    :param window_overlap_fraction: Overlap between successive windows (default 0.5)
    :param sampling_frequency: sampling frequency in Hertz, default 2.5 Hz
    :param depth:
    :param latitude:
    :param longitude:
    :return:
    """
    spectral_time, frequencies, co_spectra = estimate_co_spectra(
        epoch_time,
        (x, y, z),
        window,
        segment_length_seconds,
        window_overlap_fraction,
        sampling_frequency,
        spectral_window,
    )

    szz, a1, b1, a2, b2 = extract_moments(co_spectra, index_x=0, index_y=1, index_z=2)

    if depth is None:
        depth = numpy.full(szz.shape[0], numpy.inf)
    else:
        depth = numpy.interp(spectral_time, depth["time"].values, depth.values)

    if latitude is None:
        latitude = numpy.full(szz.shape[0], numpy.nan)
    else:
        depth = numpy.interp(spectral_time, latitude["time"].values, latitude.values)

    if longitude is None:
        longitude = numpy.full(szz.shape[0], numpy.nan)
    else:
        depth = numpy.interp(spectral_time, longitude["time"].values, longitude.values)

    dataset = Dataset(
        data_vars={
            "variance_density": (["time", "frequency"], szz),
            "a1": (["time", "frequency"], a1),
            "b1": (["time", "frequency"], b1),
            "a2": (["time", "frequency"], a2),
            "b2": (["time", "frequency"], b2),
            "depth": (["time"], depth),
            "latitude": (["time"], latitude),
            "longitude": (["time"], longitude),
        },
        coords={"time": spectral_time.astype("<M8[s]"), "frequency": frequencies},
    )
    return FrequencySpectrum(dataset)



from numba import njit
from numpy import empty_like, arctan, tan, sqrt, cos, sin, pi
from numpy.typing import NDArray


@njit(cache=True)
def integrate(time: NDArray, signal: NDArray) -> NDArray:
    coef = [3 / 8, 19 / 24, -5 / 24, 1 / 24]

    integrated_signal = empty_like(signal)
    integrated_signal[0] = 0

    # Start with Trapezoidal rule
    for ii in range(1, 3):
        dt = time[ii] - time[ii - 1]
        integrated_signal[ii] = (
            integrated_signal[ii - 1] + (signal[ii] + signal[ii - 1]) / 2 * dt
        )

    # Then apply Adams Moulton
    for ii in range(3, len(signal)):
        dt = time[ii] - time[ii - 1]
        if dt > 0.41:
            integrated_signal[ii] = 0.0
            continue

        integrated_signal[ii] = (
            integrated_signal[ii - 1]
            + coef[0] * signal[ii] * dt
            + coef[1] * signal[ii - 1] * dt
            + coef[2] * signal[ii - 2] * dt
            + coef[3] * signal[ii - 3] * dt
        )
    return integrated_signal


@njit(cache=True)
def cumulative_distance(latitudes, longitudes):
    semi_major_axis = 6378137
    semi_minor_axis = 6356752.314245
    # eccentricity - squared
    eccentricity_squared = (
        semi_major_axis**2 - semi_minor_axis**2
    ) / semi_major_axis**2

    x = empty_like(latitudes)
    y = empty_like(longitudes)
    x[0] = 0
    y[0] = 0

    for ii in range(1, len(latitudes)):
        delta_longitude = (
            ((longitudes[ii] - longitudes[ii - 1] + 180) % 360 - 180) * pi / 180
        )
        delta_latitude = (latitudes[ii] - latitudes[ii - 1]) * pi / 180

        mean_latitude = (latitudes[ii] + latitudes[ii - 1]) / 2 * pi / 180

        # reduced latitude
        reduced_latitude = arctan(sqrt(1 - eccentricity_squared) * tan(mean_latitude))

        # length of a small meridian arc
        arc_length = (
            semi_major_axis
            * (1 - eccentricity_squared)
            * (1 - eccentricity_squared * sin(mean_latitude) ** 2) ** (-3 / 2)
        )

        x[ii] = x[ii - 1] + delta_longitude * semi_major_axis * cos(reduced_latitude)
        y[ii] = y[ii - 1] + arc_length * delta_latitude
    return x, y



from typing import TypedDict
from numpy import timedelta64


class SpotterConstants(TypedDict):
    sampling_interval_gps: float
    sampling_interval_spectra: float
    sampling_interval_location: float
    number_of_samples: int
    number_of_frequencies: int
    n_channel: bool


def spotter_constants(settings="default") -> SpotterConstants:
    if settings == "default":
        return SpotterConstants(
            sampling_interval_gps=timedelta64(400, "ms"),
            sampling_interval_spectra=timedelta64(3600, "s"),
            sampling_interval_location=timedelta64(60, "s"),
            number_of_samples=256,
            number_of_frequencies=128,
            n_channel=False,
        )
    else:
        raise ValueError("unknown settings")
